#cython: language_level=3

version = (csleqp.SLEQP_VERSION_MAJOR,
           csleqp.SLEQP_VERSION_MINOR,
           csleqp.SLEQP_VERSION_PATCH)

git_branch = csleqp.SLEQP_GIT_BRANCH
git_commit_hash = csleqp.SLEQP_GIT_COMMIT_HASH
